from pathlib import Path

import pytest  # type:ignore[import]
from helpers import create_maus_and_assert, write_to_file_or_assert_equality  # type:ignore[import]


class TestAnmeldungMaus:
    """
    A unit test that ensures that the 11001/11002/11003 MAUS.json are created.
    """

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11001.json")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11001_52e(self, datafiles):
        create_maus_and_assert(
            flat_ahb_path=Path(datafiles) / "11001.json",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11001_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11002.json")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11002_52e(self, datafiles):
        create_maus_and_assert(
            flat_ahb_path=Path(datafiles) / "11002.json",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11002_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11003.json")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11003_52e(self, datafiles):
        create_maus_and_assert(
            flat_ahb_path=Path(datafiles) / "11003.json",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11003_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11004.json")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11004_52e(self, datafiles):
        create_maus_and_assert(
            flat_ahb_path=Path(datafiles) / "11004.json",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11004_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11005.json")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11005_52e(self, datafiles):
        create_maus_and_assert(
            flat_ahb_path=Path(datafiles) / "11005.json",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11005_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11006.json")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11006_52e(self, datafiles):
        create_maus_and_assert(
            flat_ahb_path=Path(datafiles) / "11006.json",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11006_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11023.json")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11023_52e(self, datafiles):
        create_maus_and_assert(
            flat_ahb_path=Path(datafiles) / "11023.json",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11023_maus.json"),
        )



from datetime import datetime, timezone
from typing import Optional, Tuple

import pytest  # type:ignore[import]

from maus.edifact import (
    EdifactFormat,
    EdifactFormatVersion,
    EdiMetaData,
    get_current_edifact_format_version,
    get_edifact_format_version,
    get_format_of_pruefidentifikator,
    is_edifact_boilerplate,
)


class TestEdifact:
    """
    Tests the edifact module
    """

    @pytest.mark.parametrize(
        "expectation_tuple",
        [
            ("11042", EdifactFormat.UTILMD),
            ("13002", EdifactFormat.MSCONS),
            ("25001", EdifactFormat.UTILTS),
        ],
    )
    def test_pruefi_to_format(self, expectation_tuple: Tuple[str, EdifactFormat]):
        """
        Tests that the prüfis can be mapped to an EDIFACT format
        """
        assert get_format_of_pruefidentifikator(expectation_tuple[0]) == expectation_tuple[1]

    @pytest.mark.parametrize(
        "key_date,expected_result",
        [
            pytest.param(
                datetime(2021, 1, 1, 0, 0, 0, tzinfo=timezone.utc),
                EdifactFormatVersion.FV2104,
                id="Anything before 2021-04-01",
            ),
            pytest.param(datetime(2021, 5, 1, 0, 0, 0, tzinfo=timezone.utc), EdifactFormatVersion.FV2104),
            pytest.param(datetime(2021, 10, 1, 0, 0, 0, tzinfo=timezone.utc), EdifactFormatVersion.FV2110),
            pytest.param(datetime(2022, 7, 1, 0, 0, 0, tzinfo=timezone.utc), EdifactFormatVersion.FV2110),
            pytest.param(datetime(2022, 10, 1, 0, 0, 0, tzinfo=timezone.utc), EdifactFormatVersion.FV2210),
            pytest.param(
                datetime(2050, 10, 1, 0, 0, 0, tzinfo=timezone.utc), EdifactFormatVersion.FV2304
            ),  # or what ever is the latest version
        ],
    )
    def test_format_version_from_keydate(self, key_date: datetime, expected_result: EdifactFormatVersion):
        actual = get_edifact_format_version(key_date)
        assert actual == expected_result

    def test_get_current_format_version(self):
        actual = get_current_edifact_format_version()
        assert isinstance(actual, EdifactFormatVersion) is True

    @pytest.mark.parametrize(
        "illegal_pruefi",
        [None, "", "asdas", "01234"],
    )
    def test_illegal_pruefis(self, illegal_pruefi: Optional[str]):
        """
        Test that illegal pruefis are not accepted
        :return:
        """
        with pytest.raises(ValueError):
            get_format_of_pruefidentifikator(illegal_pruefi)  # type:ignore[arg-type] # ok, because this raises an error

    def test_edi_meta_data_instantiation(self):
        actual = EdiMetaData(
            pruefidentifikator="11042",
            edifact_format=EdifactFormat.UTILMD,
            edifact_format_version=EdifactFormatVersion.FV2210,
        )
        assert isinstance(actual, EdiMetaData) is True

    def test_edi_meta_data_instantiation_with_default_format(self):
        actual = EdiMetaData(
            pruefidentifikator="11042",
            edifact_format_version=EdifactFormatVersion.FV2210,
        )
        assert isinstance(actual, EdiMetaData) is True
        assert actual.edifact_format == EdifactFormat.UTILMD

    def test_edi_meta_data_instantiation_with_error(self):
        with pytest.raises(ValueError) as value_error:
            _ = EdiMetaData(
                pruefidentifikator="13002",  # <-- 13002 is not a UTILMD pruefi
                edifact_format=EdifactFormat.UTILMD,
                edifact_format_version=EdifactFormatVersion.FV2210,
            )
        assert (
            f"{13002}' is incompatible with '{EdifactFormat.UTILMD}'; expected '{EdifactFormat.MSCONS}' instead"
            in str(value_error)
        )

    @pytest.mark.parametrize("pruefi", [pytest.param("10000")])
    def test_pruefi_to_format_not_mapped_exception(self, pruefi: str):
        """
        Test that pruefis that are not mapped to an edifact format are not accepted
        """
        with pytest.raises(ValueError) as excinfo:
            _ = get_format_of_pruefidentifikator(pruefi)

        assert "No Edifact format was found for pruefidentifikator" in excinfo.value.args[0]

    @pytest.mark.parametrize(
        "segment_code,expected_is_boilerplate",
        [
            pytest.param(None, True),
            pytest.param("UNH", False),
            pytest.param("UNT", True),
            pytest.param("UNZ", True),
            pytest.param("DTM", False),
        ],
    )
    def test_is_boilerplate_segment(self, segment_code: Optional[str], expected_is_boilerplate: bool):
        actual_is_boilerplate = is_edifact_boilerplate(segment_code)
        assert actual_is_boilerplate == expected_is_boilerplate



from pathlib import Path

import pytest  # type:ignore[import]

from maus.reader.mig_xml_reader import MigXmlReader, check_file_can_be_parsed_as_mig_xml

ALL_MIG_XML_FILES = pytest.mark.datafiles(
    "./migs/FV2204/template_xmls/utilmd_1131.xml",
    "./migs/FV2204/template_xmls/mscons_1154.xml",
    "./migs/FV2204/template_xmls/utilmd_1154.xml",
    "./migs/FV2204/template_xmls/utilmd_2379.xml",
    "./migs/FV2204/template_xmls/utilmd_2380.xml",
    "./migs/FV2204/template_xmls/utilmd_3055.xml",
    "./migs/FV2204/template_xmls/utilmd_7402.xml",
    "./migs/FV2204/template_xmls/utilmd_3225.xml",
    "./migs/FV2204/template_xmls/utilmd_6063.xml",
    "./migs/FV2204/template_xmls/utilmd_9013.xml",
    "./migs/FV2204/template_xmls/utilmd_6411.xml",
    "./migs/FV2204/template_xmls/reqote.xml",
    "./migs/FV2210/template_xmls/utilmd_7143.xml",
    "./migs/FV2210/template_xmls/utilmd_7059.xml",
    "./migs/FV2210/template_xmls/utilmd_7037.xml",
    "./migs/FV2210/template_xmls/utilmd_7037_z02.xml",
)


class TestMigXmlReaderRealData:
    """
    Tests the behaviour of the Message Implementation Guide model with real world example data
    """

    @ALL_MIG_XML_FILES
    def test_is_parsable(self, datafiles):
        check_file_can_be_parsed_as_mig_xml(Path(datafiles) / Path("utilmd_3225.xml"))
        # if no exception is thrown, the test is successful



from pathlib import Path

import pytest  # type:ignore[import]
from helpers import create_maus_and_assert, write_to_file_or_assert_equality  # type:ignore[import]


class TestAbmeldungMaus:
    """
    A unit test that ensures that the 11007/11008/11009 MAUS.json are created.
    """

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11007.json")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11007_52e(self, datafiles):
        create_maus_and_assert(
            flat_ahb_path=Path(datafiles) / "11007.json",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11007_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11008.json")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11008_52e(self, datafiles):
        create_maus_and_assert(
            flat_ahb_path=Path(datafiles) / "11008.json",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11008_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11009.json")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11009_52e(self, datafiles):
        create_maus_and_assert(
            flat_ahb_path=Path(datafiles) / "11009.json",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11009_maus.json"),
        )



from pathlib import Path

import pytest  # type:ignore[import]
from helpers import create_maus_and_assert, write_to_file_or_assert_equality  # type:ignore[import]


class TestStammdatenaenderungMaus:
    """
    A unit test that ensures that the 111xx MAUS.json are created.
    """

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11109.json")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11109_52e(self, datafiles):
        create_maus_and_assert(
            flat_ahb_path=Path(datafiles) / "11109.json",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11109_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11115.json")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11115_52e(self, datafiles):
        create_maus_and_assert(
            flat_ahb_path=Path(datafiles) / "11115.json",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11115_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11116.json")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11116_52e(self, datafiles):
        create_maus_and_assert(
            flat_ahb_path=Path(datafiles) / "11116.json",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11116_maus.json"),
        )

    # @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    # @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11120.json")
    # @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    # def test_maus_creation_11120_52e(self, datafiles):
    #    create_maus_and_assert(
    #        flat_ahb_path=Path(datafiles) / "11120.json",
    #         sgh_path=Path(datafiles) / "sgh_utilmd.json",
    #        template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
    #         maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11120_maus.json"),
    #     )

    # @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    # @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11124.json")
    # @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    # def test_maus_creation_11124_52e(self, datafiles):
    #     create_maus_and_assert(
    #         flat_ahb_path=Path(datafiles) / "11124.json",
    #         sgh_path=Path(datafiles) / "sgh_utilmd.json",
    #         template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
    #         maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11124_maus.json"),
    #     )
    #
    # @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    # @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11127.json")
    # @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    # def test_maus_creation_11127_52e(self, datafiles):
    #     create_maus_and_assert(
    #         flat_ahb_path=Path(datafiles) / "11127.json",
    #         sgh_path=Path(datafiles) / "sgh_utilmd.json",
    #         template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
    #         maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11127_maus.json"),
    #     )

    """
    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11117.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11117(self, datafiles):
        create_maus_and_assert(
            flat_ahb_path=Path(datafiles) / "11117.json",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11117_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11119.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11119(self, datafiles):
        create_maus_and_assert(
            flat_ahb_path=Path(datafiles) / "11119.json",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11119_maus.json"),
        )
    """

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11161.json")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11161_52e(self, datafiles):
        create_maus_and_assert(
            flat_ahb_path=Path(datafiles) / "11161.json",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11161_maus.json"),
        )

    # 11170 seems to be gone in FV2210

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11186.json")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11186_52e(self, datafiles):
        create_maus_and_assert(
            flat_ahb_path=Path(datafiles) / "11186.json",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11186_maus.json"),
        )

    # @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    # @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11123.json")
    # @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    # def test_maus_creation_11123_52e(self, datafiles):
    #     create_maus_and_assert(
    #         flat_ahb_path=Path(datafiles) / "11123.json",
    #         sgh_path=Path(datafiles) / "sgh_utilmd.json",
    #         template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
    #         maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11123_maus.json"),
    #     )
    #
    # @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    # @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11218.json")
    # @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    # def test_maus_creation_11218_52e(self, datafiles):
    #     create_maus_and_assert(
    #         flat_ahb_path=Path(datafiles) / "11218.json",
    #         sgh_path=Path(datafiles) / "sgh_utilmd.json",
    #         template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
    #         maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11218_maus.json"),
    #     )
    #
    # @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    # @pytest.mark.datafiles("../../machine-readable_anwendungshandbuecher/FV2210/UTILMD/flatahb/11219.json")
    # @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    # def test_maus_creation_11219_52e(self, datafiles):
    #     create_maus_and_assert(
    #         flat_ahb_path=Path(datafiles) / "11219.json",
    #         sgh_path=Path(datafiles) / "sgh_utilmd.json",
    #         template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
    #         maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11219_maus.json"),
    #     )

    # 11194 seems to be gone in FV2210



from typing import Dict, List, Optional, Tuple
from uuid import UUID

import pytest  # type:ignore[import]
from example_data_11042 import (  # type:ignore[import]
    example_flat_ahb_11042,
    example_sgh_11042,
    expected_changes_11042,
    expected_locations_11042,
)
from jsonpath_ng.ext import parse  # type:ignore[import] #  jsonpath is just installed in the tests
from more_itertools import last

from maus.models.anwendungshandbuch import AhbLine
from maus.models.message_implementation_guide import SegmentGroupHierarchy
from maus.navigation import (
    AhbLocation,
    AhbLocationLayer,
    _AhbLocationDistance,
    _enhance_with_next_segment,
    _enhance_with_next_value_pool_entry,
    _find_common_ancestor_from_sgh,
    _is_opening_segment_line_border,
    _PseudoAhbLocation,
    calculate_distance,
    determine_hierarchy_changes,
    determine_locations,
    find_common_ancestor,
)

sgh_utilmd_fv2204 = SegmentGroupHierarchy(
    segment_group=None,
    opening_segment="UNH",
    sub_hierarchy=[
        SegmentGroupHierarchy(segment_group="SG1", opening_segment="RFF", sub_hierarchy=[]),
        SegmentGroupHierarchy(
            segment_group="SG2",
            opening_segment="NAD",
            sub_hierarchy=[SegmentGroupHierarchy(segment_group="SG3", opening_segment="CTA", sub_hierarchy=[])],
        ),
        SegmentGroupHierarchy(
            segment_group="SG4",
            opening_segment="IDE",
            sub_hierarchy=[
                SegmentGroupHierarchy(
                    segment_group="SG5",
                    opening_segment="LOC",
                    sub_hierarchy=None,
                ),
                SegmentGroupHierarchy(segment_group="SG6", opening_segment="RFF", sub_hierarchy=None),
                SegmentGroupHierarchy(
                    segment_group="SG8",
                    opening_segment="SEQ",
                    sub_hierarchy=[
                        SegmentGroupHierarchy(segment_group="SG9", opening_segment="QTY", sub_hierarchy=None),
                        SegmentGroupHierarchy(segment_group="SG10", opening_segment="CCI", sub_hierarchy=None),
                    ],
                ),
                SegmentGroupHierarchy(segment_group="SG12", opening_segment="NAD", sub_hierarchy=None),
            ],
        ),
    ],
)


class TestNavigation:
    @pytest.mark.parametrize(
        "lines,expected",
        [
            pytest.param(
                [
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code=None,
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry=None,
                        name="Das Eine",
                        guid=None,
                    ),
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry=None,
                        name="Das Eine",
                        guid=None,
                    ),
                ],
                [
                    (
                        AhbLine(
                            segment_group_key="SG4",
                            segment_code=None,
                            ahb_expression="X",
                            data_element="0333",
                            value_pool_entry=None,
                            name="Das Eine",
                            guid=None,
                        ),
                        "FOO",
                    ),
                    (
                        AhbLine(
                            segment_group_key="SG4",
                            segment_code="FOO",
                            ahb_expression="X",
                            data_element="0333",
                            value_pool_entry=None,
                            name="Das Eine",
                            guid=None,
                        ),
                        "FOO",
                    ),
                ],
            )
        ],
    )
    def test_enhance_line_with_next_segment(self, lines: List[AhbLine], expected: List[Tuple[AhbLine, str]]):
        actual = _enhance_with_next_segment(lines)
        assert actual == expected

    @pytest.mark.parametrize(
        "lines,expected",
        [
            pytest.param(
                [
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code=None,
                        ahb_expression="X",
                        data_element=None,
                        value_pool_entry=None,
                        name="Das Eine",
                        guid=None,
                    ),
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry="Z01",
                        name="Das Eine",
                        guid=None,
                    ),
                ],
                [
                    (
                        AhbLine(
                            segment_group_key="SG4",
                            segment_code=None,
                            ahb_expression="X",
                            data_element=None,
                            value_pool_entry=None,
                            name="Das Eine",
                            guid=None,
                        ),
                        "Z01",
                    ),
                    (
                        AhbLine(
                            segment_group_key="SG4",
                            segment_code="FOO",
                            ahb_expression="X",
                            data_element="0333",
                            value_pool_entry="Z01",
                            name="Das Eine",
                            guid=None,
                        ),
                        "Z01",
                    ),
                ],
            )
        ],
    )
    def test_enhance_line_with_next_value_pool_entry(self, lines: List[AhbLine], expected: List[Tuple[AhbLine, str]]):
        actual = _enhance_with_next_value_pool_entry(lines)
        assert actual == expected

    @pytest.mark.parametrize(
        "opening_segment,this_line,next_line,next_filled_segment,expected_result",
        [
            pytest.param(
                "CCI",
                AhbLine(
                    guid=UUID("b1538d12-9bf4-47a8-b34e-b9c012cc1000"),
                    segment_group_key="SG10",
                    segment_code="CCI",
                    data_element="7037",
                    value_pool_entry="Z07",
                    name="Verbrauch",
                    ahb_expression=None,
                    section_name="Lieferrichtung",
                ),
                AhbLine(
                    guid=UUID("505273ea-2287-49dd-abdc-56320026dde6"),
                    segment_group_key="SG10",
                    segment_code=None,
                    data_element=None,
                    value_pool_entry=None,
                    name=None,
                    ahb_expression=None,
                    section_name="Zugeordnete Marktpartner",
                ),
                "CCI",
                True,
            ),
            pytest.param(
                "NAD",
                AhbLine(
                    guid=UUID("49538504-37f0-4b9b-b894-c3e8ecc64327"),
                    segment_group_key="SG12",
                    segment_code="RFF",
                    data_element="1154",
                    value_pool_entry=None,
                    name="ID der Marktlokation",
                    ahb_expression=None,
                    section_name="Referenz auf die ID der Marktlokation",
                ),
                AhbLine(
                    guid=UUID("57608e02-1eec-4e34-a77e-f6a9aa9026ff"),
                    segment_group_key="SG12",
                    segment_code=None,
                    data_element=None,
                    value_pool_entry=None,
                    name=None,
                    ahb_expression=None,
                    section_name="Korrespondenzanschrift des Kunden des Lieferanten",
                ),
                "NAD",
                True,
            ),
            pytest.param(
                "CCI",
                AhbLine(
                    guid=UUID("7cc3c40b-07d4-457a-a60f-1661485adcac"),
                    segment_group_key="SG10",
                    segment_code="CAV",
                    data_element="1131",
                    value_pool_entry="MP-ID",
                    name=None,
                    ahb_expression=None,
                    section_name="Lieferant",
                ),
                AhbLine(
                    guid=UUID("0e774e0c-7714-4853-82c8-742e65362eae"),
                    segment_group_key="SG10",
                    segment_code=None,
                    data_element=None,
                    value_pool_entry=None,
                    name=None,
                    ahb_expression=None,
                    section_name="Übertragungsnetzbetreiber",
                ),
                "CAV",
                False,
            ),
        ],
    )
    def test_is_opening_segment_line_border(
        self,
        opening_segment: str,
        this_line: AhbLine,
        next_line: AhbLine,
        next_filled_segment: Optional[str],
        expected_result: bool,
    ):
        actual = _is_opening_segment_line_border(this_line, next_line, next_filled_segment, opening_segment)
        assert actual == expected_result

    @pytest.mark.parametrize(
        "location_x, location_y, x_is_sub_of_y, y_is_parent_of_x",
        [
            pytest.param(
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD")
                    ]
                ),
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD")
                    ]
                ),
                True,
                True,
                id="equality",
            ),
            pytest.param(
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD")
                    ]
                ),
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG1", opening_segment_code="BAR", opening_qualifier="QWE")
                    ]
                ),
                False,
                False,
                id="not equality",
            ),
            pytest.param(
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD"),
                        AhbLocationLayer(segment_group_key="SG1", opening_segment_code="BAR", opening_qualifier="QWE"),
                    ]
                ),
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD"),
                    ]
                ),
                True,
                True,
                id="true sub",
            ),
        ],
    )
    def test_ahb_position_is_sub_or_parent_of(
        self, location_x: AhbLocation, location_y: AhbLocation, x_is_sub_of_y: bool, y_is_parent_of_x: bool
    ):
        assert location_x.is_sub_location_of(location_y) == x_is_sub_of_y
        assert location_y.is_parent_of(location_x) == y_is_parent_of_x

    @pytest.mark.parametrize(
        "location_x, location_y, expected",
        [
            pytest.param(
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD"),
                        AhbLocationLayer(segment_group_key="SG1", opening_segment_code="BAR", opening_qualifier="QWE"),
                    ]
                ),
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD"),
                        AhbLocationLayer(segment_group_key="SG1", opening_segment_code="BAR", opening_qualifier="QWE"),
                    ]
                ),
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD"),
                        AhbLocationLayer(segment_group_key="SG1", opening_segment_code="BAR", opening_qualifier="QWE"),
                    ]
                ),
                id="x == y",
            ),
            pytest.param(
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD"),
                        AhbLocationLayer(segment_group_key="SG1", opening_segment_code="BAR", opening_qualifier="QWE"),
                    ]
                ),
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD"),
                    ]
                ),
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD"),
                    ]
                ),
                id="x is parent of y",
            ),
            pytest.param(
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD"),
                        AhbLocationLayer(segment_group_key="SG1", opening_segment_code="BAR", opening_qualifier="QWE"),
                    ]
                ),
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD"),
                        AhbLocationLayer(segment_group_key="SG1", opening_segment_code="BAR", opening_qualifier="IOP"),
                    ]
                ),
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD"),
                    ]
                ),
                id="x and y are siblings",
            ),
        ],
    )
    def test_find_common_ancestor(self, location_x: AhbLocation, location_y: AhbLocation, expected: AhbLocation):
        actual = find_common_ancestor(location_x, location_y)
        assert actual == expected

    @pytest.mark.parametrize(
        "sgh, sg_x, sg_y, expected",
        [
            pytest.param(
                sgh_utilmd_fv2204,
                "SG2",
                "SG2",
                _PseudoAhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key=None, opening_segment_code="UNH", opening_qualifier=None),
                        AhbLocationLayer(segment_group_key="SG2", opening_segment_code="NAD", opening_qualifier=None),
                    ]
                ),
                id="x and y are siblings (S2, UNH)",
            ),
            pytest.param(
                sgh_utilmd_fv2204,
                "SG12",
                "SG6",
                _PseudoAhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key=None, opening_segment_code="UNH", opening_qualifier=None),
                        AhbLocationLayer(segment_group_key="SG4", opening_segment_code="IDE", opening_qualifier=None),
                    ]
                ),
                id="x and y are siblings (S6.12)",
            ),
            pytest.param(
                sgh_utilmd_fv2204,
                "SG3",
                "SG12",
                _PseudoAhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key=None, opening_segment_code="UNH", opening_qualifier=None),
                    ]
                ),
                id="deep nested",
            ),
            pytest.param(
                sgh_utilmd_fv2204,
                "SG9",
                "SG6",
                _PseudoAhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key=None, opening_segment_code="UNH", opening_qualifier=None),
                        AhbLocationLayer(segment_group_key="SG4", opening_segment_code="IDE", opening_qualifier=None),
                    ]
                ),
                id="deep nested 2",
            ),
        ],
    )
    def test_find_common_ancestor_from_sgh(
        self, sg_x: str, sg_y: str, sgh: SegmentGroupHierarchy, expected: _PseudoAhbLocation
    ):
        actual = _find_common_ancestor_from_sgh(sg_x, sg_y, sgh)
        assert actual == expected

    @pytest.mark.parametrize(
        "location_x, location_y",
        [
            pytest.param(
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD"),
                        AhbLocationLayer(segment_group_key="SG1", opening_segment_code="BAR", opening_qualifier="QWE"),
                    ]
                ),
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG7", opening_segment_code="FOO", opening_qualifier="ASD"),
                        AhbLocationLayer(segment_group_key="SG1", opening_segment_code="BAR", opening_qualifier="QWE"),
                    ]
                ),
                id="there is no common root",
            ),
        ],
    )
    def test_find_no_common_ancestor(self, location_x: AhbLocation, location_y: AhbLocation):
        with pytest.raises(ValueError):
            _ = find_common_ancestor(location_x, location_y)

    @pytest.mark.parametrize(
        "location_x, location_y, expected",
        [
            pytest.param(
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD"),
                        AhbLocationLayer(segment_group_key="SG1", opening_segment_code="BAR", opening_qualifier="QWE"),
                    ]
                ),
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD"),
                        AhbLocationLayer(segment_group_key="SG1", opening_segment_code="BAR", opening_qualifier="QWE"),
                    ]
                ),
                _AhbLocationDistance(layers_down=0, layers_up=0),
                id="x == y",
            ),
            pytest.param(
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD"),
                        AhbLocationLayer(segment_group_key="SG1", opening_segment_code="BAR", opening_qualifier="QWE"),
                    ]
                ),
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD"),
                    ]
                ),
                _AhbLocationDistance(layers_down=0, layers_up=1),
                id="x is parent of y",
            ),
            pytest.param(
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD"),
                        AhbLocationLayer(segment_group_key="SG1", opening_segment_code="BAR", opening_qualifier="QWE"),
                    ]
                ),
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key="SG0", opening_segment_code="FOO", opening_qualifier="ASD"),
                        AhbLocationLayer(segment_group_key="SG1", opening_segment_code="BAR", opening_qualifier="IOP"),
                    ]
                ),
                _AhbLocationDistance(layers_down=1, layers_up=1),
                id="x and y are siblings",
            ),
        ],
    )
    def test_calculate_distance(self, location_x: AhbLocation, location_y: AhbLocation, expected: _AhbLocationDistance):
        actual = calculate_distance(location_x, location_y)
        assert actual == expected

    def _assert_consistency(self, locations: List[AhbLocation]) -> None:
        """
        assert that the locations are self-consistent in that regard, that the same segment group is always opened
        with the same segment code (which is derived from the SegmentGroupHierarchy)
        """
        sg_opening_segments: Dict[str, str] = {}  # maps SG-keys to their respective opening segments, e.g. "SG6":"RFF"
        for location in locations:
            for layer in location.layers:
                if layer.segment_group_key in sg_opening_segments:
                    if layer.segment_group_key is not None:
                        assert layer.opening_segment_code == sg_opening_segments[layer.segment_group_key]
                elif layer.segment_group_key is not None:
                    sg_opening_segments[layer.segment_group_key] = layer.opening_segment_code

    def test_determine_locations(self):
        actual_locations = [
            x[1] for x in determine_locations(example_sgh_11042, ahb_lines=example_flat_ahb_11042.lines)
        ]
        self._assert_consistency(actual_locations)
        assert actual_locations == expected_locations_11042

    def test_determine_differential_changes(self):
        actual_locations = [
            x[1] for x in determine_locations(example_sgh_11042, ahb_lines=example_flat_ahb_11042.lines)
        ]
        actual_changes = [x[1] for x in determine_hierarchy_changes(example_flat_ahb_11042.lines, example_sgh_11042)]
        location_and_changes = [
            (line, loc, change)
            for line, loc, change in zip(example_flat_ahb_11042.lines, actual_locations, actual_changes)
        ]
        assert location_and_changes is not None  # you may use this to update the 10k lines of equality assertions
        if actual_changes != expected_changes_11042:
            for n, (actual, expected) in enumerate(zip(actual_changes, expected_changes_11042, strict=True)):
                assert actual == expected, f"Error at line {n}: {example_flat_ahb_11042.lines[n]}"
        assert actual_changes == expected_changes_11042



import pytest  # type:ignore[import]
from helpers import create_maus_and_assert, write_to_file_or_assert_equality  # type:ignore[import]


class TestGeraeteuebernahmeMaus:
    """
    A unit test that ensures that the 11057/58/59 MAUS.json are created
    """



from typing import List, Literal, Optional

import pytest  # type:ignore[import]
from lxml import etree  # type:ignore[import]

from maus.reader.etree_element_helpers import get_ahb_name_or_none, get_nested_qualifiers, get_segment_group_key_or_none


def _string_to_element(xml_elem: str) -> etree.Element:
    """
    helper method that adds an XML preamble to the given string, parses the concatenated string as etree
    and returns the root element
    """
    valid_xml = '<?xml version="1.0"?>' + xml_elem
    root = etree.fromstring(valid_xml)
    return root


def _string_to_elements(xml_elem: str) -> List[etree.Element]:
    """
    helper method that adds an XML preamble and virtual root to the given string,
    parses the concatenated string as etree and returns the elements on root level
    """
    valid_xml = '<?xml version="1.0"?><virtual_root>' + xml_elem + "</virtual_root>"
    root = etree.fromstring(valid_xml)
    return [x for x in root.findall("./")]


class TestEtreeSingleElementHelpers:
    """
    Tests the behaviour of static helper methods (for single elements)
    """

    @pytest.mark.parametrize(
        "xml_string, expected",
        [
            pytest.param('<foo ref="SG14"></foo>', "SG14"),
            pytest.param("<bar></bar>", None),
        ],
    )
    def test_get_segment_group_key_or_none(self, xml_string: str, expected: Optional[str]):
        element = _string_to_element(xml_string)
        assert get_segment_group_key_or_none(element) == expected

    @pytest.mark.parametrize(
        "xml_string, expected",
        [
            pytest.param('<foo ahbName="X"></foo>', "X"),
            pytest.param("<bar></bar>", None),
        ],
    )
    def test_get_ahb_name_or_none(self, xml_string: str, expected: Optional[str]):
        element = _string_to_element(xml_string)
        assert get_ahb_name_or_none(element) == expected

    @pytest.mark.parametrize(
        "ref_or_key, xml_string, expected",
        [
            pytest.param("ref", '<foo ref="hello"></foo>', None),
            pytest.param("ref", '<bar ref="NAD:2:0[1:0=MS]"></bar>', ["MS"]),
            pytest.param("ref", '<baz ref="RFF:1:1[RFF:1:0=Z13]"></baz>', ["Z13"]),
            pytest.param("ref", '<baz ref="DTM:1:1[1:0=469]"></baz>', ["469"]),
            pytest.param("key", '<baz asd="hello"></baz>', None),
            pytest.param("key", '<foo key="NAD:2:0[1:0=MS]"></foo>', ["MS"]),
            pytest.param("key", '<foo key="QTY:1:1[1:0=265|1:0=Z10|1:0=Z08]"></foo>', ["265", "Z10", "Z08"]),
            pytest.param("ref", '<field ref="FTX:4:(0,4)[1:0=ACB]" />', ["ACB"]),
            pytest.param("ref", '<field ref="CAV:1:0[CCI:1:0=Z02|CCI:1:0=Z04]" />', ["Z02", "Z04"]),
            pytest.param("key", '<class key="NAD:5:(0,3)[NAD:1:0=Z08]" ref="SG12" />', ["Z08"]),
            pytest.param("key", '<class key="NAD:4:(0,4)[NAD:1:0=Z07]" ref="SG12" />', ["Z07"]),
        ],
    )
    def test_get_nested_qualifier(self, ref_or_key: Literal["ref", "key"], xml_string: str, expected: List[str]):
        element = _string_to_element(xml_string)
        assert get_nested_qualifiers(ref_or_key, element) == expected



from typing import List

import pytest  # type:ignore[import]
from unit_tests.serialization_test_helper import assert_serialization_roundtrip  # type:ignore[import]

from maus.mig_ahb_matching import merge_lines_with_same_data_element
from maus.models.anwendungshandbuch import AhbLine
from maus.models.edifact_components import (
    DataElement,
    DataElementFreeText,
    DataElementValuePool,
    EdifactStack,
    ValuePoolEntry,
)


class TestMaus:
    """
    Tests the behaviour of the MAUS.
    """

    @pytest.mark.parametrize(
        "ahb_lines, expected_data_element",
        [
            pytest.param(
                # The following two lines should be merged into a single data element value pool because they
                # refer to the same data element but have different value pool entries.
                [
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry="E01",
                        name="Das Eine",
                        guid=None,
                    ),
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry="E02",
                        name="Das Andere",
                        guid=None,
                    ),
                ],
                DataElementValuePool(
                    discriminator="$",
                    value_pool=[
                        ValuePoolEntry(
                            qualifier="E01",
                            meaning="Das Eine",
                            ahb_expression="X",
                        ),
                        ValuePoolEntry(
                            qualifier="E02",
                            meaning="Das Andere",
                            ahb_expression="X",
                        ),
                    ],
                    data_element_id="0333",
                    entered_input=None,
                ),
            ),
            pytest.param(
                # The following single lines should result in a single DataElement with free text option
                [
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry=None,
                        name="Das Eine",
                        guid=None,
                    ),
                ],
                DataElementFreeText(
                    discriminator="$",
                    ahb_expression="X",
                    entered_input=None,
                    data_element_id="0333",
                ),
            ),
        ],
    )
    def test_data_element_merging(self, ahb_lines: List[AhbLine], expected_data_element: DataElement):
        actual = merge_lines_with_same_data_element(ahb_lines, EdifactStack(levels=[]))
        assert actual == expected_data_element

    @pytest.mark.parametrize(
        "ahb_lines",
        [
            pytest.param(
                [
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry="E01",
                        name="Das Eine",
                        guid=None,
                    ),
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0334",
                        value_pool_entry="E02",
                        name="Das Andere",
                        guid=None,
                    ),
                ],
                id="lines with different data element",
            ),
            pytest.param([], id="empty list"),
        ],
    )
    def test_data_element_value_error(self, ahb_lines: List[AhbLine]):
        with pytest.raises(ValueError):
            merge_lines_with_same_data_element(ahb_lines, EdifactStack(levels=[]))

    @pytest.mark.parametrize(
        "ahb_lines, expected_data_element",
        [
            pytest.param(
                # The following two lines should be merged into a single data element value pool because they
                # refer to the same data element but have different value pool entries.
                [
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry="E01",
                        name="Das Eine",
                        guid=None,
                    ),
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry="E02",
                        name="Das Andere",
                        guid=None,
                    ),
                ],
                DataElementValuePool(
                    discriminator="$",
                    value_pool=[
                        ValuePoolEntry(
                            qualifier="E01",
                            meaning="Das Eine",
                            ahb_expression="X",
                        ),
                        ValuePoolEntry(
                            qualifier="E02",
                            meaning="Das Andere",
                            ahb_expression="X",
                        ),
                    ],
                    entered_input=None,
                    data_element_id="0333",
                ),
            ),
            pytest.param(
                # The following single lines should result in a single DataElement with free text option
                [
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry=None,
                        name="Das Eine",
                        guid=None,
                    ),
                ],
                DataElementFreeText(
                    discriminator="$",
                    ahb_expression="X",
                    entered_input=None,
                    data_element_id="0333",
                ),
            ),
        ],
    )
    def test_nest_segment_groups_into_each_other(self, ahb_lines: List[AhbLine], expected_data_element: DataElement):
        actual = merge_lines_with_same_data_element(ahb_lines, EdifactStack(levels=[]))
        assert actual == expected_data_element



from typing import Tuple

import pytest  # type:ignore[import]
from lxml import etree  # type:ignore[import]
from lxml.etree import Element  # type:ignore[import]

from maus.models.edifact_components import EdifactStack, EdifactStackLevel
from maus.navigation import AhbLocation, AhbLocationLayer
from maus.reader.mig_xml_reader import MigXmlReader


class TestMigXmlReaderMwe:
    """
    Tests the behaviour of the Message Implementation Guide model with manually crafted minimal working examples.
    """

    # todo: you could probably use an indirect parametrization instead of the static helper methods below

    @staticmethod
    def _prepare_xml_reader_and_element(xml_string: str, element_xpath: str) -> Tuple[MigXmlReader, Element]:
        """
        A helper method for easier test setup
        """
        reader = MigXmlReader(xml_string)
        elements = reader._original_tree.xpath(element_xpath)
        if len(elements) != 1:
            raise ValueError(
                f"In this test you're thought to provide an xpath that matches exactly one element of the given XML but found {len(elements)}."
            )
        element = elements[0]
        return reader, element

    @staticmethod
    def _prepare_xml_reader_and_expected_element(xml_string: str, element_xpath: str) -> Tuple[MigXmlReader, Element]:
        """
        A helper method for easier test setup
        """
        reader = MigXmlReader(xml_string)
        elements = reader._original_tree.xpath(element_xpath)
        if len(elements) != 1:
            raise ValueError(
                f"In this test you're thought to provide an xpath that matches exactly one element of the given XML but found {len(elements)} for xpath '{element_xpath}'."
                # this means, the paths should be absolute
            )
        element = elements[0]
        return reader, element

    @pytest.mark.parametrize(
        "xml_string, expected_format",
        [
            pytest.param('<?xml version="1.0"?><MSCONS></MSCONS>', "MSCONS"),
            pytest.param('<?xml version="1.0"?><UTILMD></UTILMD>', "UTILMD"),
        ],
    )
    def test_get_edifact_format(self, xml_string: str, expected_format: str):
        reader = MigXmlReader(xml_string)
        assert reader.get_format_name() == expected_format

    @pytest.mark.parametrize(
        "xml_string, element_xpath, use_sanitized, expected_result",
        [
            pytest.param(
                '<?xml version="1.0"?><MSCONS><foo name="asd"><bar ahbName="qwe"/></foo></MSCONS>',
                "//MSCONS/foo/bar",
                False,
                EdifactStack.from_json_path(r'$["asd"]["qwe"]'),
            ),
            pytest.param(
                '<?xml version="1.0"?><MSCONS><class name="asd"><bar ahbName="qwe"/></class></MSCONS>',
                "//MSCONS/class/bar",
                False,
                EdifactStack(
                    levels=[
                        EdifactStackLevel(name="asd", is_groupable=True),
                        EdifactStackLevel(name="qwe", is_groupable=False),
                    ]
                ),
            ),
            pytest.param(
                '<?xml version="1.0"?><MSCONS><class ahbName="asd"><bar name="qwe"/><class ahbName="opi"/></class></MSCONS>',
                "//MSCONS/class/class",
                False,
                EdifactStack(
                    levels=[
                        EdifactStackLevel(name="asd", is_groupable=True),
                        EdifactStackLevel(name="opi", is_groupable=True),
                    ]
                ),
            ),
            pytest.param(
                '<?xml version="1.0"?><UTILMD><class name="Dokument" ref="/" key="UNB:5:0"><class name="Absender_NAD" migName="MP-ID Absender" max="1" ref="SG2" key="NAD:2:0[1:0=MS]" ahbName="MP-ID Absender" meta.type="group" meta.objType="Marktpartner"></class></class></UTILMD>',
                "//UTILMD/class/class",
                False,
                EdifactStack(
                    levels=[
                        EdifactStackLevel(name="Dokument", is_groupable=True),
                        EdifactStackLevel(name="MP-ID Absender", is_groupable=True),
                        # "MP-ID Absender" shall be preferred over "Absender_NAD"
                    ]
                ),
            ),
            pytest.param(
                '<?xml version="1.0"?><UTILMD><class name="Dokument" ref="/" key="UNB:5:0"><class name="Nachricht" ref="UNH" key="UNH:1:0" max="9999" meta.type="group"><field name="Nachrichten-Referenznummer" ref="UNH:1:0" meta.id="0062" /><field name="Kategorie" migName="Beginn der Nachricht" meta.id="1001" ref="BGM:1:0" groupBy="Beginn der Nachricht" meta.help="Hier wird die Kategorie der gesamten Nachricht für alle Vorgänge angegeben." meta.type="repository" meta.typeInfo="Kategorierepository"/></class></class></UTILMD>',
                "//UTILMD/class/class/field[2]",
                False,
                EdifactStack(
                    levels=[
                        EdifactStackLevel(name="Dokument", is_groupable=True),
                        EdifactStackLevel(name="Nachricht", is_groupable=True),
                        EdifactStackLevel(name="Kategorie", is_groupable=False),
                    ]
                ),
            ),
        ],
    )
    def test_element_to_edifact_stack(
        self, xml_string: str, element_xpath: str, use_sanitized: bool, expected_result: EdifactStack
    ):
        reader, element = TestMigXmlReaderMwe._prepare_xml_reader_and_element(xml_string, element_xpath)
        actual = reader.element_to_edifact_stack(element, use_sanitized_tree=use_sanitized)
        assert actual == expected_result

    @pytest.mark.parametrize(
        "xml_string,ahb_location,expected_result_xpath",
        [
            pytest.param(
                '<?xml version="1.0"?><MSCONS><class ref="/"><class ref="UNH"><class ref="SG1"><foo/></class></class></class></MSCONS>',
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key=None, opening_segment_code="UNH", opening_qualifier=None),
                        AhbLocationLayer(segment_group_key="SG1", opening_segment_code="FOO", opening_qualifier="BAR"),
                    ]
                ),
                "/MSCONS/class/class/class",
                id="simple unique result",
            ),
            pytest.param(
                '<?xml version="1.0"?><UTILMD><class ref="/"><class ref="UNH"><class ref="SG4" key="IDE:2:0"><class name="Netznutzungsabrechnungsdaten der Marktlokation" ref="SG8" key="SEQ:1:0[SEQ:1:0=Z45]"><foo/></class><class name="Weitere Abrechnungsdaten der Marktlokation" ref="SG8" key="SEQ:1:0[SEQ:1:0=Z46]"><field name="ID" ref="PIA:2:0[PIA:1:0=Z02]" meta.id="7140" /></class></class></class></class></UTILMD>',
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key=None, opening_segment_code="UNH", opening_qualifier=None),
                        AhbLocationLayer(segment_group_key="SG4", opening_segment_code="IDE", opening_qualifier="24"),
                        AhbLocationLayer(segment_group_key="SG8", opening_segment_code="SEQ", opening_qualifier="Z46"),
                    ],
                    data_element_id="7140",
                ),
                "/UTILMD/class/class/class/class[2]/field",
                id="distinguish by sg key, then by data element",
            ),
            pytest.param(
                '<?xml version="1.0"?><UTILMD><class ref="/"><class ref="UNH"><class ref="SG4" key="IDE:2:0"><field name="Vertragsbeginn" ref="DTM:1:1[1:0=92]" meta.format="102" meta.id="2380" meta.type="date-time" ahbName="Beginn zum" /><field name="Vertragsende" ref="DTM:1:1[1:0=93]" meta.format="102" meta.id="2380" ahbName="Ende zum" meta.type="date-time" /></class></class></class></UTILMD>',
                AhbLocation(
                    layers=[
                        AhbLocationLayer(segment_group_key=None, opening_segment_code="UNH", opening_qualifier=None),
                        AhbLocationLayer(segment_group_key="SG4", opening_segment_code="IDE", opening_qualifier="24"),
                    ],
                    data_element_id="2380",
                    qualifier="93",
                ),
                "/UTILMD/class/class/class/field[2]",
                id="DTM92 and DTM93",
            ),
        ],
    )
    def test_get_segment_group_element(self, xml_string: str, ahb_location: AhbLocation, expected_result_xpath: str):
        reader, expected_element = TestMigXmlReaderMwe._prepare_xml_reader_and_expected_element(
            xml_string, expected_result_xpath
        )
        actual = reader.get_element(ahb_location=ahb_location)
        assert actual == expected_element
